from django.apps import AppConfig


class SimpleOauth2Config(AppConfig):
    """Configuration for the simple_oauth2 app."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "simple_oauth2"



from rest_framework import serializers


class TokenSerializer(serializers.Serializer):
    """Serializer for the token endpoint input data."""

    provider = serializers.CharField()
    code = serializers.CharField()
    state = serializers.CharField()



class SimpleOAuth2Error(Exception):
    """Base exception for simple_oauth2's exceptions."""


class UnknownProvider(SimpleOAuth2Error):
    """Raised when an unknown OAuth2 provider is referenced."""

    def __init__(self, provider: str):
        self.provider = provider
        super().__init__(f"Unknown OAuth2 provider '{provider}'")



import secrets
import ssl
import string
import urllib.parse
from typing import Any

import jwt
from django.contrib.auth import get_user_model
from django.db import models
from rest_framework.exceptions import ValidationError

from simple_oauth2.settings import OAuth2ProviderSettings


def generate_nonce(size: int = 128) -> str:
    """Generate a random nonce as an hexadecimal string."""
    return secrets.token_hex(size // 2)


def generate_state(size: int = 128) -> str:
    """Generate a random state as an hexadecimal string."""
    return secrets.token_hex(size // 2)


def generate_code_verifier(size: int = 128) -> str:
    """Generate a random code verifier."""
    if not (43 <= size <= 128):
        raise ValueError("code_verifier must be 43..128 chars (RFC 7636).")
    return "".join(secrets.choice(string.ascii_letters + string.digits + "-._~") for _ in range(size))


def decode_jwt(provider: OAuth2ProviderSettings, token: str) -> dict:
    """Decode a JWT token using the provider's JKWS URI."""
    try:
        ssl_context = ssl._create_unverified_context() if not provider.VERIFY_SSL else None  # nosec: B323
        jwks_client = jwt.PyJWKClient(provider.jwks_uri(), ssl_context=ssl_context, timeout=provider.TIMEOUT)
        key = jwks_client.get_signing_key_from_jwt(token).key
    except jwt.PyJWKClientConnectionError as e:  # pragma: no cover
        raise ValidationError({"__all__": f"Failed to fetch JKWS from the provider: {e}"})
    return jwt.decode(token, key, algorithms=provider.SIGNING_ALGORITHMS, audience=provider.CLIENT_ID)


def get_user(provider: OAuth2ProviderSettings, userinfo: dict, **kwargs: Any) -> models.Model:
    """Create or update a user from the id_token and the userinfo dictionary."""
    from simple_oauth2.models import Sub

    User = get_user_model()

    claims = decode_jwt(provider=provider, token=kwargs["id_token"]) if "id_token" in kwargs else {}
    sub_value = claims.get("sub") or userinfo.get("sub")
    username = (claims.get("preferred_username") or userinfo.get("preferred_username")) or (
        claims.get("email") or userinfo.get("email")
    )
    try:
        sub = Sub.objects.get(_provider=provider.alias, sub=sub_value)
        sub.save(update_fields=["last_login"])  # Update the 'last_login' timestamp
        user = sub.user
    except Sub.DoesNotExist:
        user = User.objects.create_user(username=username)
        Sub.objects.create(user=user, _provider=provider.alias, sub=sub_value)

    user.first_name = claims.get("given_name", "") or userinfo.get("given_name", "")
    user.last_name = claims.get("family_name", "") or userinfo.get("family_name", "")
    user.email = claims.get("email", "") or userinfo.get("email", "")
    user.save()

    return user


def simple_jwt_authenticate(
    provider: OAuth2ProviderSettings, oauth2_tokens: dict[str, str], user: models.Model
) -> dict:
    """Create a simple JWT token payload."""
    from rest_framework_simplejwt.tokens import RefreshToken

    params = {
        "client_id": provider.CLIENT_ID,
        "id_token_hint": oauth2_tokens["id_token"],
        "post_logout_redirect_uri": provider.POST_LOGOUT_REDIRECT_URI,
    }
    url = urllib.parse.urljoin(provider.BASE_URL, provider.LOGOUT_PATH)
    refresh = RefreshToken.for_user(user)
    return {
        "api": {"refresh": str(refresh), "access": str(refresh.access_token)},
        "provider": oauth2_tokens | {"logout_url": f"{url}?{urllib.parse.urlencode(params)}"},
    }



import pytest
from django.contrib.auth import get_user_model

from simple_oauth2.models import Sub
from simple_oauth2.settings import oauth2_settings
from simple_oauth2.utils import generate_code_verifier, get_user
from tests.conftest import SIMPLE_OAUTH2_SETTINGS, override_oauth2_settings


def test_generate_code_verifier():
    assert generate_code_verifier()
    with pytest.raises(ValueError):
        generate_code_verifier(10)
    with pytest.raises(ValueError):
        generate_code_verifier(200)


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_get_user():
    user = get_user(
        provider=oauth2_settings["no-pkce"],
        userinfo={
            "sub": "123456",
            "preferred_username": "user",
            "email": "test@test.com",
        },
    )
    assert Sub.objects.filter(sub="123456").exists()
    assert user.username == "user"
    assert user.email == "test@test.com"


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_get_user_through_sub():
    created = get_user_model().objects.create(username="user1", email="test@test.com")
    Sub.objects.create(user=created, _provider="no-pkce", sub="123456")
    got = get_user(
        provider=oauth2_settings["no-pkce"],
        userinfo={
            "sub": "123456",
            "preferred_username": "user2",  # Use another username to ensure sub is used
            "email": "test@test.com",
        },
    )
    assert created.pk == got.pk



from django.urls import include, path
from rest_framework import routers

from simple_oauth2 import views

app_name = "simple_oauth2"

router = routers.SimpleRouter()
router.register(r"oauth2", views.OAuth2ViewSet, basename="oauth2")

urlpatterns = [
    path("", include(router.urls)),
]



import logging
import urllib.parse
from typing import Any, Callable

import requests
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

logger = logging.getLogger(__name__)

# Default settings for OAuth2 providers
DEFAULTS = {
    "OPENID_CONFIGURATION_PATH": ".well-known/openid-configuration",
    "AUTHORIZATION_SESSION_LIFETIME": 300,  # 5 minutes
    "AUTHORIZATION_EXTRA_PARAMETERS": {},
    "TOKEN_USERINFO_HANDLER": "simple_oauth2.utils.get_user",
    "TOKEN_PAYLOAD_HANDLER": "simple_oauth2.utils.simple_jwt_authenticate",
    "CODE_CHALLENGE_METHOD": "S256",
    "SCOPES": ["openid", "profile", "email"],
    "USE_PKCE": True,
    "VERIFY_SSL": True,
    "TIMEOUT": 5,
    "ALLOW_REDIRECTS": True,
}

# Mapping of settings to OpenID Connect configuration keys
CONFIGURATION_KEY = {
    "AUTHORIZATION_PATH": "authorization_endpoint",
    "TOKEN_PATH": "token_endpoint",
    "USERINFO_PATH": "userinfo_endpoint",
    "LOGOUT_PATH": "end_session_endpoint",
    "JWKS_PATH": "jwks_uri",
    "SIGNING_ALGORITHMS": "id_token_signing_alg_values_supported",
}

# Settings that may be imported from strings
IMPORT_STRINGS = {"TOKEN_USERINFO_HANDLER", "TOKEN_PAYLOAD_HANDLER"}

# Mandatory settings that must be either loaded from OpenID configuration,
# or provided by the user
MANDATORY = {
    "CLIENT_ID",
    "CLIENT_SECRET",
    "REDIRECT_URI",
    "POST_LOGOUT_REDIRECT_URI",
    "BASE_URL",
    "AUTHORIZATION_PATH",
    "TOKEN_PATH",
    "USERINFO_PATH",
    "JWKS_PATH",
    "LOGOUT_PATH",
    "SIGNING_ALGORITHMS",
}


def import_from_string(v: str, provider: str, setting_name: str) -> type:
    """Attempt to import a class from a string representation."""
    try:
        return import_string(v)
    except ImportError as e:  # pragma: no cover
        raise ImportError(
            f"Could not import {v} for SIMPLE_OAUTH2 setting '{provider}[{setting_name}]' {e.__class__.__name__}: {e}."
        )


class OAuth2ProviderSettings:
    """
    A settings object, that allows OAuth2 Provider settings to be accessed as properties.

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, alias: str, user_settings: dict):
        self._alias = alias

        # Try retrieving some settings from the provider's OpenID configuration
        user_settings = DEFAULTS | user_settings
        url = urllib.parse.urljoin(user_settings["BASE_URL"], user_settings["OPENID_CONFIGURATION_PATH"])
        provider_settings = self._load_settings_from_provider(
            url,
            user_settings["TIMEOUT"],
            user_settings["VERIFY_SSL"],
            user_settings["ALLOW_REDIRECTS"],
        )

        settings = provider_settings | user_settings
        if missing := MANDATORY - {k for k, v in settings.items() if v}:  # pragma: no cover
            raise ImproperlyConfigured(
                f"OAuth2 provider '{self.alias}' is missing mandatory settings: {', '.join(missing)}"
            )

        self._user_settings = settings

    def __getattr__(self, attr: str) -> Any:
        """Return the setting value or raise an AttributeError."""
        if attr not in self._user_settings:  # pragma: no cover
            raise AttributeError(f"Invalid SIMPLE_OAUTH2 setting: '{self.alias}[{attr}]'")
        if attr in IMPORT_STRINGS:
            return self._perform_import(self._user_settings[attr], attr)
        return self._user_settings[attr]

    def _load_settings_from_provider(
        self, url: str, timeout: int, verify: bool, allow_redirects: bool
    ) -> dict[str, Any]:
        """Load settings from the provider's OpenID configuration endpoint."""
        try:
            response = requests.get(url, timeout=timeout, verify=verify, allow_redirects=allow_redirects)
            response.raise_for_status()
        except requests.RequestException as e:  # pragma: no cover
            logger.warning(
                "Could not fetch '%s' OpenID configuration from '%s': %s",
                self.alias,
                url,
                e,
            )
            return {}
        configuration = response.json()
        return {setting: configuration[path] for setting, path in CONFIGURATION_KEY.items()}

    def _perform_import(self, value: str, setting_name: str) -> Callable:  # pragma: no cover
        """Import a class from a string representation."""
        if isinstance(value, str):
            return import_from_string(value, self.alias, setting_name)
        elif isinstance(value, Callable):
            return value
        raise ImproperlyConfigured(
            f"SIMPLE_OAUTH2 setting '{self.alias}[{setting_name}]' must be a string or callable."
        )

    @property
    def alias(self) -> str:
        """Return the provider alias."""
        return self._alias

    def jwks_uri(self) -> str:
        """Return the full JWKS URL."""
        return urllib.parse.urljoin(self.BASE_URL, self.JWKS_PATH)

    def authorization_uri(self) -> str:
        """Return the full authorization URL."""
        return urllib.parse.urljoin(self.BASE_URL, self.AUTHORIZATION_PATH)

    def token_uri(self) -> str:
        """Return the full token URL."""
        return urllib.parse.urljoin(self.BASE_URL, self.TOKEN_PATH)

    def userinfo_uri(self) -> str:
        """Return the full token URL."""
        return urllib.parse.urljoin(self.BASE_URL, self.USERINFO_PATH)


oauth2_settings = {
    alias: OAuth2ProviderSettings(alias, settings) for alias, settings in getattr(settings, "SIMPLE_OAUTH2", {}).items()
}



# Generated by Django 4.1.13 on 2025-09-23 02:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("simple_oauth2", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="session",
            name="status",
            field=models.CharField(
                choices=[
                    ("pending", "Pending"),
                    ("token_failed", "Token Failed"),
                    ("userinfo_failed", "Userinfo Failed"),
                    ("expired", "Expired"),
                    ("completed", "Completed"),
                ],
                default="pending",
                max_length=20,
            ),
        ),
    ]



import base64
import datetime
import hashlib
import urllib.parse
from typing import Any

import requests
from django.contrib.auth import get_user_model
from django.core.exceptions import NON_FIELD_ERRORS
from django.core.validators import RegexValidator
from django.db import IntegrityError, models, transaction
from django.utils import timezone
from rest_framework.exceptions import ValidationError

from simple_oauth2 import utils
from simple_oauth2.enums import Status
from simple_oauth2.exceptions import SimpleOAuth2Error, UnknownProvider
from simple_oauth2.settings import OAuth2ProviderSettings, oauth2_settings

PKCE_VALIDATOR = RegexValidator(
    regex=r"^[A-Za-z0-9\-._~]{43,128}$",
    message="code_verifier must be 43–128 chars of unreserved URL characters (RFC 7636).",
)


class Session(models.Model):
    """Represent an OAuth2 authorization session."""

    _provider = models.CharField(max_length=255, db_column="provider")

    nonce = models.CharField(max_length=255, default=utils.generate_nonce)
    state = models.CharField(max_length=255, default=utils.generate_state)
    code_verifier = models.CharField(
        max_length=128,
        default=utils.generate_code_verifier,
        validators=(PKCE_VALIDATOR,),
    )

    status = models.CharField(max_length=20, choices=Status.choices, default=Status.PENDING)
    created_at = models.DateTimeField(auto_now_add=True)
    completed_at = models.DateTimeField(null=True)

    class Meta:
        constraints = (models.UniqueConstraint(fields=("_provider", "state"), name="session_provider_state_key"),)
        indexes = (models.Index(fields=("created_at",)),)

    @classmethod
    def start(cls, provider: str) -> "Session":
        """Try creating a unique Session while avoiding race condition."""
        if provider not in oauth2_settings:
            raise UnknownProvider(provider)
        for _ in range(10):
            try:
                with transaction.atomic():
                    return Session.objects.create(_provider=provider)
            except IntegrityError:
                continue
        raise SimpleOAuth2Error("Could not create a unique authorization session after 10 attempts.")

    @property
    def provider(self) -> OAuth2ProviderSettings:
        """Return the provider configuration associated with this Session."""
        if self._provider in oauth2_settings:
            return oauth2_settings[self._provider]
        raise UnknownProvider(self._provider)

    @property
    def use_pkce(self) -> bool:
        """Return whether or not the Session uses PKCE."""
        return self.provider.USE_PKCE

    @property
    def code_challenge_method(self) -> str:
        """Return the code challenge method associated with this Session."""
        return self.provider.CODE_CHALLENGE_METHOD

    @property
    def code_challenge(self) -> str:
        """Return the code challenge from code_verifier and code_challenge_method."""
        if self.code_challenge_method.lower() == "plain":
            return self.code_verifier

        elif self.code_challenge_method.lower() == "s256":
            digest = hashlib.sha256(self.code_verifier.encode("ascii")).digest()
            return base64.urlsafe_b64encode(digest).rstrip(b"=").decode("ascii")

        raise SimpleOAuth2Error(f"Unknown challenge method: '{self.code_challenge_method}'")

    def has_expired(self) -> bool:
        """Return whether the authorization session as expired."""
        now = timezone.now().astimezone(datetime.timezone.utc)
        if timezone.is_naive(self.created_at):
            ts = timezone.make_aware(self.created_at, datetime.timezone.utc)
        else:
            ts = self.created_at.astimezone(datetime.timezone.utc)
        if (now - ts).total_seconds() > self.provider.AUTHORIZATION_SESSION_LIFETIME:
            return True
        return False

    def authentication_url(self) -> str:
        """Generate the OAuth2 authentication URL for the given alias."""
        params = {
            "response_type": "code",
            "client_id": self.provider.CLIENT_ID,
            "scope": " ".join(self.provider.SCOPES),
            "nonce": self.nonce,
            "state": self.state,
            "redirect_uri": self.provider.REDIRECT_URI,
        }
        if self.use_pkce:
            params |= {
                "code_challenge": self.code_challenge,
                "code_challenge_method": self.provider.CODE_CHALLENGE_METHOD,
            }
        for key, value in self.provider.AUTHORIZATION_EXTRA_PARAMETERS.items():
            params[key] = value
        return f"{self.provider.authorization_uri()}?{urllib.parse.urlencode(params)}"

    def get_tokens(self, code: str) -> dict[str, str]:
        """Exchange the authorization code for tokens."""
        if self.has_expired():
            self.status = Status.EXPIRED
            self.save(update_fields=["status"])
            raise ValidationError({NON_FIELD_ERRORS: "Authorization session has expired."})

        data = {
            "grant_type": "authorization_code",
            "client_id": self.provider.CLIENT_ID,
            "client_secret": self.provider.CLIENT_SECRET,
            "code": code,
            "redirect_uri": self.provider.REDIRECT_URI,
        }
        if self.use_pkce:
            data["code_verifier"] = self.code_verifier

        try:
            response = requests.post(
                self.provider.token_uri(),
                data=data,
                timeout=self.provider.TIMEOUT,
                verify=self.provider.VERIFY_SSL,
                allow_redirects=self.provider.ALLOW_REDIRECTS,
            )
            response.raise_for_status()
        except requests.RequestException as e:
            self.status = Status.TOKEN_FAILED
            self.save(update_fields=("status",))
            raise ValidationError(
                {"__all__": f"Failed to retrieve tokens from provider: {e.response.content.decode()}"}
            )

        return response.json()

    def get_user(self, access_token: str, **kwargs: Any) -> models.Model:
        """Fetch the user infos from the provider and feet it to the handler."""
        try:
            response = requests.get(
                self.provider.userinfo_uri(),
                headers={"Authorization": f"Bearer {access_token}"},
                timeout=self.provider.TIMEOUT,
                verify=self.provider.VERIFY_SSL,
                allow_redirects=self.provider.ALLOW_REDIRECTS,
            )
            response.raise_for_status()
        except requests.RequestException as e:
            self.status = Status.USERINFO_FAILED
            self.save(update_fields=("status",))
            raise ValidationError(
                {"__all__": f"Failed to fetch user info from the provider: {e.response.content.decode()}."}
            )

        user = self.provider.TOKEN_USERINFO_HANDLER(self.provider, response.json(), **kwargs)

        return user

    def get_payload(self, oauth2_tokens: dict[str, str], user: models.Model) -> dict:
        """Fetch the payload from the provider and feed it to the handler."""
        payload = self.provider.TOKEN_PAYLOAD_HANDLER(self.provider, oauth2_tokens, user)
        self.status = Status.COMPLETED
        self.completed_at = timezone.now()
        self.save(update_fields=("status", "completed_at"))
        return payload


class Sub(models.Model):
    """Link a user to a given OAuth2 provider and its sub."""

    _provider = models.CharField(max_length=255, db_column="provider")
    sub = models.CharField(max_length=255)
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)
    last_login = models.DateTimeField(auto_now=True)

    class Meta:
        constraints = (models.UniqueConstraint(fields=("_provider", "sub"), name="sub_provider_sub_key"),)
        indexes = (models.Index(fields=("sub",)),)



# Generated by Django 4.1.13 on 2025-09-08 04:51

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import simple_oauth2.utils


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Session",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("_provider", models.CharField(db_column="provider", max_length=255)),
                (
                    "nonce",
                    models.CharField(default=simple_oauth2.utils.generate_nonce, max_length=255),
                ),
                (
                    "state",
                    models.CharField(default=simple_oauth2.utils.generate_state, max_length=255),
                ),
                (
                    "code_verifier",
                    models.CharField(
                        default=simple_oauth2.utils.generate_code_verifier,
                        max_length=128,
                        validators=[
                            django.core.validators.RegexValidator(
                                message="code_verifier must be 43–128 chars of unreserved URL characters (RFC 7636).",
                                regex="^[A-Za-z0-9\\-._~]{43,128}$",
                            )
                        ],
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PENDING", "Pending"),
                            ("TOKEN_FAILED", "Token Failed"),
                            ("USERINFO_FAILED", "Userinfo Failed"),
                            ("EXPIRED", "Expired"),
                            ("COMPLETED", "Completed"),
                        ],
                        default="PENDING",
                        max_length=20,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("completed_at", models.DateTimeField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Sub",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("_provider", models.CharField(db_column="provider", max_length=255)),
                ("sub", models.CharField(max_length=255)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("last_login", models.DateTimeField(auto_now=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddIndex(
            model_name="session",
            index=models.Index(fields=["created_at"], name="simple_oaut_created_7c172d_idx"),
        ),
        migrations.AddConstraint(
            model_name="session",
            constraint=models.UniqueConstraint(fields=("_provider", "state"), name="session_provider_state_key"),
        ),
        migrations.AddIndex(
            model_name="sub",
            index=models.Index(fields=["sub"], name="simple_oaut_sub_7eb942_idx"),
        ),
        migrations.AddConstraint(
            model_name="sub",
            constraint=models.UniqueConstraint(fields=("_provider", "sub"), name="sub_provider_sub_key"),
        ),
    ]



from datetime import timedelta
from types import SimpleNamespace
from unittest.mock import patch
from urllib.parse import parse_qs, urlparse

import pytest
from django.urls import reverse
from django.utils import timezone
from rest_framework.test import APIClient

from simple_oauth2.exceptions import SimpleOAuth2Error
from simple_oauth2.models import Session
from simple_oauth2.settings import oauth2_settings
from tests.conftest import SIMPLE_OAUTH2_SETTINGS, override_oauth2_settings


@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_url_no_provider():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"))
    assert response.status_code == 400
    assert response.json() == {
        "provider": f"OAuth2 provider required, allowed values are: '{', '.join(oauth2_settings.keys())}'"
    }


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_url_unknown_provider():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "unknown"})
    assert response.status_code == 400
    assert response.json() == {
        "provider": f"Unknown OAuth2 provider 'unknown', allowed values are: '{', '.join(oauth2_settings.keys())}'"
    }


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_url_pkce_unknown_alg():
    client = APIClient()
    with pytest.raises(SimpleOAuth2Error, match="Unknown challenge method: 'unknown'"):
        client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "pkce-unknown-alg"})


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_extra_params_url():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "extra-params"})
    assert response.status_code == 200, response.json()
    assert "url" in response.json()
    url = response.json()["url"]
    query_params = parse_qs(urlparse(url).query)
    assert query_params["client_id"] == ["extra-params"]
    assert query_params["response_type"] == ["code"]
    assert query_params["redirect_uri"] == ["https://example.com/callback"]
    assert query_params["scope"] == ["openid profile email"]
    assert "state" in query_params
    assert "nonce" in query_params
    assert "foo" in query_params
    assert query_params["foo"] == ["bar"]
    assert "baz" in query_params
    assert query_params["baz"] == ["qux"]


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_token_unknown_session():
    client = APIClient()
    response = client.post(
        reverse("simple_oauth2:oauth2-token"),
        {"provider": "pkce", "state": "unknown", "code": "code"},
    )
    assert response.status_code == 400
    assert response.json() == {"state": "No ongoing session matches the provided state."}


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_token_expired():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "no-pkce"})
    state = parse_qs(urlparse(response.json()["url"]).query)["state"][0]
    Session.objects.filter(state=state).update(created_at=timezone.now() - timedelta(days=1))
    response = client.post(
        reverse("simple_oauth2:oauth2-token"),
        {"provider": "no-pkce", "state": state, "code": "no-pkce-code"},
    )
    assert response.status_code == 400
    assert response.json() == {"__all__": "Authorization session has expired."}


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_token_fails():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "token-fails"})
    state = parse_qs(urlparse(response.json()["url"]).query)["state"][0]
    response = client.post(
        reverse("simple_oauth2:oauth2-token"),
        {"provider": "token-fails", "state": state, "code": "token-fails-code"},
    )
    assert response.status_code == 400
    assert response.json() == {"__all__": 'Failed to retrieve tokens from provider: {"detail": "error"}'}


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_userinfo_fails():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "userinfo-fails"})
    state = parse_qs(urlparse(response.json()["url"]).query)["state"][0]
    response = client.post(
        reverse("simple_oauth2:oauth2-token"),
        {"provider": "userinfo-fails", "state": state, "code": "userinfo-fails-code"},
    )
    assert response.status_code == 400
    assert response.json() == {"__all__": 'Failed to fetch user info from the provider: {"detail": "error"}.'}


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_without_pkce_url():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "no-pkce"})
    assert response.status_code == 200
    assert "url" in response.json()
    url = response.json()["url"]
    query_params = parse_qs(urlparse(url).query)
    assert query_params["client_id"] == ["no-pkce"]
    assert query_params["response_type"] == ["code"]
    assert query_params["redirect_uri"] == ["https://example.com/callback"]
    assert query_params["scope"] == ["openid profile email"]
    assert "state" in query_params
    assert "nonce" in query_params
    assert "code_challenge" not in query_params
    assert "code_challenge_method" not in query_params


@pytest.mark.django_db
@patch(
    "jwt.jwks_client.PyJWKClient.get_signing_key",
    side_effect=lambda token: SimpleNamespace(key="key"),
)
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_without_pkce_token():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "no-pkce"})
    state = parse_qs(urlparse(response.json()["url"]).query)["state"][0]
    response = client.post(
        reverse("simple_oauth2:oauth2-token"),
        {"provider": "no-pkce", "state": state, "code": "no-pkce-code"},
    )
    assert response.status_code == 200, response.json()
    data = response.json()
    assert "api" in data
    assert "access" in data["api"]
    assert "refresh" in data["api"]
    assert "provider" in data
    assert "logout_url" in data["provider"]
    assert "id_token" in data["provider"]
    assert "access_token" in data["provider"]


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_with_pkce_plain_url():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "pkce-plain"})
    assert response.status_code == 200
    assert "url" in response.json()
    url = response.json()["url"]
    query_params = parse_qs(urlparse(url).query)
    assert query_params["client_id"] == ["pkce-plain"]
    assert query_params["response_type"] == ["code"]
    assert query_params["redirect_uri"] == ["https://example.com/callback"]
    assert query_params["scope"] == ["openid profile email"]
    assert "state" in query_params
    assert "nonce" in query_params
    assert "code_challenge" in query_params
    assert "code_challenge_method" in query_params
    assert query_params["code_challenge_method"] == ["plain"]


@pytest.mark.django_db
@patch(
    "jwt.jwks_client.PyJWKClient.get_signing_key",
    side_effect=lambda token: SimpleNamespace(key="key"),
)
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_with_pkce_plain_token():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "pkce-plain"})
    state = parse_qs(urlparse(response.json()["url"]).query)["state"][0]
    response = client.post(
        reverse("simple_oauth2:oauth2-token"),
        {"provider": "pkce-plain", "state": state, "code": "pkce-plain-code"},
    )
    assert response.status_code == 200, response.json()
    data = response.json()
    assert "api" in data
    assert "access" in data["api"]
    assert "refresh" in data["api"]
    assert "provider" in data
    assert "logout_url" in data["provider"]
    assert "id_token" in data["provider"]
    assert "access_token" in data["provider"]


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_with_pkce_s256_url():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "pkce-s256"})
    assert response.status_code == 200
    assert "url" in response.json()
    url = response.json()["url"]
    query_params = parse_qs(urlparse(url).query)
    assert query_params["client_id"] == ["pkce-s256"]
    assert query_params["response_type"] == ["code"]
    assert query_params["redirect_uri"] == ["https://example.com/callback"]
    assert query_params["scope"] == ["openid profile email"]
    assert "state" in query_params
    assert "nonce" in query_params
    assert "code_challenge" in query_params
    assert "code_challenge_method" in query_params
    assert query_params["code_challenge_method"] == ["s256"]


@pytest.mark.django_db
@patch(
    "jwt.jwks_client.PyJWKClient.get_signing_key",
    side_effect=lambda token: SimpleNamespace(key="key"),
)
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_oauth2_with_pkce_s256_token():
    client = APIClient()
    response = client.get(reverse("simple_oauth2:oauth2-url"), {"provider": "pkce-s256"})
    state = parse_qs(urlparse(response.json()["url"]).query)["state"][0]
    response = client.post(
        reverse("simple_oauth2:oauth2-token"),
        {"provider": "pkce-s256", "state": state, "code": "pkce-s256-code"},
    )
    assert response.status_code == 200, response.json()
    data = response.json()
    assert "api" in data
    assert "access" in data["api"]
    assert "refresh" in data["api"]
    assert "provider" in data
    assert "logout_url" in data["provider"]
    assert "id_token" in data["provider"]
    assert "access_token" in data["provider"]



from django.db import models


class Status(models.TextChoices):
    """Valid status values for a Session."""

    PENDING = "pending"
    TOKEN_FAILED = "token_failed"
    USERINFO_FAILED = "userinfo_failed"
    EXPIRED = "expired"
    COMPLETED = "completed"



from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.request import Request
from rest_framework.response import Response

from simple_oauth2.enums import Status
from simple_oauth2.exceptions import UnknownProvider
from simple_oauth2.models import Session
from simple_oauth2.serializers import TokenSerializer
from simple_oauth2.settings import oauth2_settings


class OAuth2ViewSet(viewsets.GenericViewSet):
    """Allows authentication through FranceConnect."""

    authentication_classes: list[type] = []
    permission_classes: list[type] = []
    serializer_class = TokenSerializer

    @action(detail=False, methods=("get",))
    def url(self, request: Request) -> Response:
        """Return the OAuth2 URL the login button must use."""
        if (provider := request.query_params.get("provider")) is None:
            raise ValidationError(
                {"provider": f"OAuth2 provider required, allowed values are: '{', '.join(oauth2_settings)}'"}
            )
        try:
            session = Session.start(provider)
        except UnknownProvider:
            raise ValidationError(
                {
                    "provider": f"Unknown OAuth2 provider '{provider}', allowed values are: '{', '.join(oauth2_settings)}'"
                }
            )
        return Response({"url": session.authentication_url()})

    @action(detail=False, methods=("post",), serializer_class=TokenSerializer)
    def token(self, request: Request) -> Response:
        """Return tokens using the code obtained from your provider."""
        serializer = self.get_serializer(data=self.request.data)
        serializer.is_valid(raise_exception=True)

        try:
            session = Session.objects.get(
                _provider=serializer.validated_data["provider"],
                state=serializer.validated_data["state"],
                status=Status.PENDING,
            )
        except Session.DoesNotExist:
            raise ValidationError({"state": "No ongoing session matches the provided state."})

        oauth2_tokens = session.get_tokens(serializer.validated_data["code"])
        user = session.get_user(**oauth2_tokens)
        payload = session.get_payload(oauth2_tokens, user)
        return Response(payload)



from contextlib import contextmanager
from datetime import timedelta
from typing import Any, Callable, TypeVar

import pytest
from django.db import models
from django.utils import timezone

from simple_oauth2.exceptions import SimpleOAuth2Error, UnknownProvider
from simple_oauth2.models import Session
from simple_oauth2.utils import generate_state
from tests.conftest import SIMPLE_OAUTH2_SETTINGS, override_oauth2_settings

STATE1 = generate_state()
STATE2 = generate_state()

T = TypeVar("T")


@contextmanager
def override_field_default(model: type[models.Model], field_name: str, default: Callable[[], Any]):
    field = model._meta.get_field(field_name)
    old = field.default
    try:
        field._get_default = default
        field.default = default
        yield
    finally:
        field.default = old
        field._get_default = old


class ValuesDefault:
    """Callable class to return values from a list one after the other."""

    def __init__(self, values: list[T]):
        self.values = values
        self.index = 0

    def __call__(self) -> T:
        value = self.values[self.index]
        self.index = (self.index + 1) % len(self.values)
        return value


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_start_multiple_try():
    with override_field_default(Session, "state", ValuesDefault([STATE1] * 5 + [STATE2])):
        session = Session.start("no-pkce")
        assert session.state == STATE1
        session2 = Session.start("no-pkce")
        assert session2.state == STATE2


@pytest.mark.django_db
@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_start_integrity_error():
    with override_field_default(Session, "state", ValuesDefault([STATE1] * 11)):
        Session.start("no-pkce")
        with pytest.raises(
            SimpleOAuth2Error,
            match="Could not create a unique authorization session after 10 attempts.",
        ):
            Session.start("no-pkce")


@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_provider_unknown():
    session = Session(_provider="unknown")
    with pytest.raises(UnknownProvider, match="Unknown OAuth2 provider 'unknown'"):
        _ = session.provider


@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_has_expired():
    session = Session(_provider="no-pkce", created_at=timezone.now())
    assert not session.has_expired()
    session.created_at = timezone.now() - timedelta(days=1)
    assert session.has_expired()


@override_oauth2_settings(SIMPLE_OAUTH2_SETTINGS)
def test_has_expired_naive():
    session = Session(_provider="no-pkce", created_at=timezone.now().replace(tzinfo=None))
    assert not session.has_expired()
    session.created_at = (timezone.now() - timedelta(days=1)).replace(tzinfo=None)
    assert session.has_expired()
